import _device as dev
from latent_class_mixed_model import LatentClassMixedModel
from latent_class_model import LatentClassModel
from mixed_logit import MixedLogit
from multinomial_logit import MultinomialLogit
from search import Search
from siman import*
from threshold import*
from misc import*



"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
MISCELLANEOUS CODES
"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""

import numpy as np

''' -------------------------------------------------------------------- '''
''' Function. Create a list with 'val' replicated n times                '''
''' -------------------------------------------------------------------- '''
def make_list(val, n):
    return [val] * n

def list_of_zeros(n): # {
    return make_list(0, n)
# }

###########################################################################
# CODES COMMON TO LATENT_CLASS AND LATENT_CLASS_MIXED
###########################################################################

''' -------------------------------------------------------------------- '''
''' Function. Define as [None, None, ..., None] where |[...]| = num_classes '''
''' -------------------------------------------------------------------- '''
def initialise_avail_latent(avail_latent, num_classes):
#{
    # CONSIDER REPLACING 'None' WITH 'NaN'
    avail_latent = np.repeat(None, num_classes) if avail_latent is None else avail_latent
    return avail_latent
# }

''' -------------------------------------------------------------------- '''
''' Function. Initialise fit_intercept flag                              '''
''' -------------------------------------------------------------------- '''
def initialise_fit_intercept(params, fit_inter = None, is_latent = False):
# {
    if is_latent:
    # {
        if '_inter' in params:
            return True
        else:
            return False
    # }
    if fit_inter is not None:
    # {
        if '_inter' in params:
            return params
        return '_inter' in np.concatenate(params)
    # }
    else:
    # {
        for i in params:
            if '_inter' in i:
                return True                    
        return False
    # }
# }

''' -------------------------------------------------------------------- '''
''' Function. Initialise class_params_spec                               '''
''' If unspecified, default to using all varnames in each latent class   '''
''' -------------------------------------------------------------------- '''
def initialise_class_params_spec(class_params_spec, isvars, varnames, num_classes):
# {
    if class_params_spec is None: # {
        if isvars is not None:
            class_vars = [var for var in varnames if var not in isvars]
        else:
            class_vars = varnames

        # Replicate class_vars and vstack the replicates
        class_params_spec = np.vstack(make_list(class_vars, num_classes))
    # }
    return class_params_spec
# }

''' -------------------------------------------------------------------- '''
''' Function.                                                            '''
''' -------------------------------------------------------------------- '''
def initialise_membership_as_probability(member_params_spec):
# {
    membership_as_probability = True if member_params_spec is True else False
    return membership_as_probability
# }

''' -------------------------------------------------------------------- '''
''' Function. Initialise member_params_spec                              '''
''' -------------------------------------------------------------------- '''
def initialise_member_params_spec(membership_as_probability, member_params_spec, isvars, varnames, num_classes):
# {
    if membership_as_probability:
        member_params_spec = np.vstack(make_list('dummy', num_classes-1))  # Replicate 'dummy' and vstack the replicates

    if member_params_spec is None:
    # {
        if isvars is not None: member_vars = isvars
        else: member_vars = varnames

        # Replicate member_vars and vstack the replicates
        member_params_spec = np.vstack(make_list(member_vars, num_classes-1))
    # }
    return member_params_spec
# }

''' -------------------------------------------------------------------- '''
''' Function. Initialise intercept_opts                                  '''
''' -------------------------------------------------------------------- '''
def initialise_opts(intercept_opts, num_classes):
# {
    if intercept_opts is None:
        intercept_opts = {}
    else:
    # {
        if 'class_intercept_alts' in intercept_opts:  # {
            if len(intercept_opts['class_intercept_alts']) != num_classes:
                raise ValueError("The key class_intercept_alts in intercept_opts must be the same length as num_classes")
    # }
    return intercept_opts
# }
''' -------------------------------------------------------------------- '''
''' Function. Setup multinomial_logit or mixed_logit                     '''
''' -------------------------------------------------------------------- '''
def setup_logit(i, logit, X, y, varnames, class_params_spec, avail, alts, transvars, gtol,
    mxl=False, panels=None, randvars=None, correlated_vars=None, n_draws=None, mnl_init=None):
# {
    varnames_i, fit_intercept_i = class_params_spec[i], False
    if '_inter' in varnames_i:  
    # {
        # X = np.hstack((np.ones((X.shape[0], 1)), X))
        # if '_inter' not in varnames: #FIXME why does the interpept miss
         #   varnames = np.append(['_inter'], varnames)
        fit_intercept_i = True #this was true I am just adding it to the data
        varnames_i = [var for var in varnames_i]
        varnames_i = [var for var in varnames_i if var != '_inter']
        #varnames_i = [var for var in varnames_i if var != '_inter']           
    # }                                                                

    X_i_idx = [ii for ii, name in enumerate(varnames) if name in varnames_i]
    X_i = np.array(X)[:, X_i_idx]

    transvars_i = []
    if transvars is not None:
        transvars_i = [transvar for transvar in transvars if transvar in varnames_i]

    if mxl:
    # {
        randvars_i = {k: v for k, v in randvars.items() if k in varnames_i}
        correlated_vars_i = correlated_vars
        if correlated_vars is not None and isinstance(correlated_vars, list):
            correlated_vars_i = [corvar for corvar in correlated_vars if corvar in varnames_i]

        logit.setup(X_i, y, varnames_i, alts, avail=avail, transvars=transvars_i,
                randvars=randvars_i, panels=panels, fit_intercept=fit_intercept_i,
                correlated_vars=correlated_vars_i, n_draws=n_draws, mnl_init=mnl_init, gtol=gtol)
    # }
    else:
        logit.setup(X_i, y, varnames_i, alts, avail=avail, transvars=transvars_i,
              fit_intercept=fit_intercept_i, gtol=gtol)
    logit.fit()
    return logit
# }

''' -------------------------------------------------------------------- '''
''' Function.                                                            '''
''' -------------------------------------------------------------------- '''
def revise_betas(i, logit, betas, intercept_opts, alts):
# {
    coefs = logit.coeff_est + np.random.normal(0, 0.01, len(logit.coeff_est))  # Add some random noise
    if 'class_intercept_alts' in intercept_opts:
    # {
        intercept_idx = np.array(intercept_opts['class_intercept_alts'][i]) - 2
        J = len(np.unique(alts))  # Code shortcut
        coefs = np.concatenate((coefs[intercept_idx], coefs[(J - 1):]))
    # }
    betas[i] = coefs + np.random.normal(0, 0.01, len(coefs))  # Add small amounts of noise
    return betas
# }



"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
IMPLEMENTATION: HARMONY SEARCH
"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""

"""
BACKGROUND - HARMONY SEARCH

Initialization: The algorithm starts by initializing a population of candidate solutions, called "harmonies." 
Each harmony represents a potential solution to the optimization problem.

Improvisation: Similar to musicians trying out different combinations of notes, Harmony Search generates new 
solutions by combining elements from existing harmonies. This is done through a process called "harmony memory 
consideration."

Evaluation: Each newly generated harmony is evaluated based on a fitness function, which measures how 
good the solution is in terms of solving the optimization problem.

Updating Harmony Memory: The best harmonies are selected to update the harmony memory, replacing the worst 
harmonies if they are better.

Memory Consideration: During the improvisation process, the algorithm considers both the current harmonies 
and the memory of the best solutions found so far to guide the search towards better solutions.

Pitch Adjustment: Similar to how musicians adjust their notes to achieve harmony, Harmony Search introduces 
randomness by adjusting certain elements (parameters) of the harmonies. This helps in exploring different
regions of the search space.

Termination: The process continues for a certain number of iterations or until a termination criterion is 
met (e.g., a satisfactory solution is found).

"""

''' ---------------------------------------------------------- '''
''' LIBRARIES                                                  '''
''' ---------------------------------------------------------- '''
from search import*
import matplotlib.pyplot as plt
import datetime
import math
import pandas as pd

''' ---------------------------------------------------------- '''
''' CONSTANTS                                                  '''
''' ---------------------------------------------------------- '''

sol_keys = ['asvars', 'isvars', 'randvars', 'bcvars', 'corvars', 'bctrans', 'cor']

''' ---------------------------------------------------------- '''
''' CLASS FOR HARMONY SEARCH (HS)                              '''
''' ---------------------------------------------------------- '''
class HarmonySearch(Search):
# {
    """ Docstring """

    # ===================
    # CLASS PARAMETERS
    # ===================

    """
    int mem_size: Harmony memory size / Defaults to 10.
    int min_classes: Minimum number of latent classes. Defaults to 1
    int max_classes: Maximum number of latent classes. Defaults to 5
    float min_harm: Minimum harmony memory consideration rate / Defaults to 0.6.
    float max_harm: Maximum harmony memory consideration rate / Defaults to 0.9.
    float max_pitch: Maximum pitch adjustment rate / Defaults to 0.85
    float min_pitch: Minimum pitch adjustment / Defaults to 0.3
    int maxiter: Maximum iteratioms / Defaults to 30.
    float prop_local: Proportion of iterations without local search /  Defaults to 0.8.
    int threshold: Threshold to compare new solution with worst solution / Defaults to 15

    bool termination_override: termination flag that overrides the default / Defaults to False.
        If true, the search will run for each number of latent classes
        between min_classes and max_classes

    iter_prop: Proportion of maxiter after which local search is initiated / float
    """

    # =======================
    # CLASS FUNCTIONS
    # =======================

    """
    1. set_control_parameters()
    2. create_opposite_solution(self, sol);
    3. initialize_memory(self, nb_sols);
    4. build_solution(self, memory, prop);
    5. pitch_adjustment(self, sol, pitch);
    6. get_best_features(self, memory);
    7. local_search(self, improved_harmony, iter, pitch);
    8. improvise(self);
    9. run(self, latent=False);
    10. sort_memory(self, mem);
    11. insert_solution(self, solution):
    """

    ''' ---------------------------------------------------------- '''
    ''' Function. Constructor                                      '''
    ''' ---------------------------------------------------------- '''
    def set_control_parameters(self, max_harm=0.9, min_harm=0.6, max_pitch=0.85, min_pitch=0.3,
        max_mem=10, maxiter=30, threshold=15, prop_local=0.8, generate_plots=True):
    # {
        self.max_harm = max_harm  # Maximum Harmony Memory Considering Rate / float
        self.min_harm = min_harm  # Minimum Harmony Memory Considering Rate / float
        self.max_pitch = max_pitch  # Maximum Pitch Adjusting Rate / float
        self.min_pitch = min_pitch  # Minimum Pitch Adjusting Rate / float
        self.max_mem = max_mem  # Harmony memory size / int
        self.maxiter = maxiter  # Maximum number of iterations / int
        self.threshold = threshold  # Convergence threshold /float
        self.prop_local = prop_local  # Proportion of maxiter
        self.perform_local = int(self.prop_local * self.maxiter)  # When to apply local search
        self.generate_plots = generate_plots
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Constructor                                      '''
    ''' ---------------------------------------------------------- '''
    def __init__(self, param: Parameters):
    # {
        super().__init__(param) # Call base class constructor
        self.set_control_parameters() # Assume default options - hence no inputs here
        self.pitch = self.max_pitch
        self.memory = []  # List of solutions is empty
        self.all_solutions = [] # Avoid generating same solution twice

        self.results_file = open("results.txt", "w")  # File to output results
        self.progress_file = open("progress.txt", "w")  # File to output convergence information
# }

    ''' ---------------------------------------------------------------- '''
    ''' Function                                                         '''
    ''' ---------------------------------------------------------------- '''
    def sort_memory(self, mem):
    # {
        if self.param.nb_crit > 1:
            mem = self.non_dominant_sorting(mem)
        else:
            mem = sorted(mem, key=lambda sol: sol.obj[0])
        return mem
    # }

    ''' ---------------------------------------------------------------- '''
    ''' Function                                                         '''
    ''' ---------------------------------------------------------------- '''
    def create_opposite_solution(self, sol):
    # {
        opp_sol = self.generate_solution()
        for key in sol_keys:  # Iterate through variable types
        # {
            skip = (not opp_sol[key] or isinstance(opp_sol[key], bool) or getattr(self.param, 'ps_' + key))
            if skip:
                continue  # Skip current loop
            else:
            # {
                opp_sol[key] = [v for v in opp_sol[key] if v not in sol[key]]  # Filter out elements in sol[key]
                if self.param.ps_intercept is None:
                    opp_sol['asc_ind'] = not sol['asc_ind']
            # }
        # }
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        if self.param.avail_rvars:
            opp_sol['randvars'] = {k: self.param.generator.choice(self.param.distr)
                                   for k in opp_sol['randvars'] if k in opp_sol['asvars']}
        else:
            opp_sol['randvars'] = {}
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        opp_sol['corvars'] = [corvar for corvar in opp_sol['corvars'] if corvar in opp_sol['randvars']]
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        if self.param.avail_bcvars:
            opp_sol['bcvars'] = [bcvar for bcvar in opp_sol['bcvars']
                                 if bcvar in opp_sol['asvars'] and bcvar not in opp_sol['corvars']]
        else:
            opp_sol['bcvars'] = []
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        self.revise_solution('class_params_spec', opp_sol, sol)
        self.revise_solution('member_params_spec', opp_sol, sol)
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # }

    ''' ---------------------------------------------------------------- '''
    ''' Function. Initialization of harmony search memory                '''
    ''' ---------------------------------------------------------------- '''
    def initialize_memory(self, nb_sols):
    # {
        """ This function initializes the harmony memory and opposite
        harmony memory with unique random solutions. The harmony memory
        stores initial solutions, while the opposite harmony memory
        stores solutions that include variables not included in the
        harmony memory. If the generated solution converges, it's added
        to the harmony memory. Otherwise, the function generates an
        "opposite" solution and, if it converges, adds it to the
        opposite harmony memory.
        """

        mem, opp_mem = [], []
        for counter in range(30000):
        # {
            sol = self.generate_solution()  # Generated solution
            sol, converged = self.evaluate_solution(sol)
            if converged: mem.append(sol) # Add new solution to list

            # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            # Create opposite solution that has non_included variables
            # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            opp_sol = self.create_opposite_solution(sol)
            opp_sol, converged = self.evaluate_solution(opp_sol)
            if converged: opp_mem.append(opp_sol)
            # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

            mem += opp_mem    # Aggregate solutions
            mem = get_unique(mem, 0)  # Keep unique solutions only - Compare by first objective

            # QUERY: WHY NOT FILTER BY sol.obj[1] AS WELL?

            mem = [sol for sol in mem if abs(sol.obj[0]) < BOUND] # Filter out poor solutions
            if len(mem) >= nb_sols:
                return mem[:nb_sols]  # Exit and return list of solutions
        # }
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        return mem # Failed to generate required number of solutions
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Build new solution using Harmony Memory          '''
    ''' A new solution, could either be built from an existing one '''
    ''' or constructed randomly.                                   '''
    ''' ---------------------------------------------------------- '''
    def build_solution(self, memory, prop):
    # {
        """ This function decides whether to build a new solution from an existing solution
        in the harmony memory or to generate a completely new solution, based on a random number and the
        Harmony Memory Consideration Rate (HMCR). If the random number is less than or equal to prop,
        it selects a proportion of the features from a randomly chosen existing solution to build the new solution.
        Otherwise, it generates a completely new solution """


        bin = [0,1] # Binary values
        prob = [1-prop, prop]  # Range
        new_sol = Solution(nb_crit=self.nb_crit)    # Create a new solution object

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # IS THIS NECESSARY?
        '''fronts, pareto = None, None
        if nb_crit > 1: # {
            memory = self.non_dominant_sorting(memory)
            fronts = self.get_fronts(memory)
            pareto = self.get_pareto(fronts, memory)
        # }'''
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        if self.param.generator.rand() > prop:
            new_sol = self.generate_solution()  # Generate a new solution
        else:
        # {
            choice = self.param.generator.choice(len(memory)) # Choose one of the member solutions
            chosen_sol = memory[choice] # Define reference to the chosen member solution

            # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            # OPTIONAL CODE.
            # size = len(chosen_sol['asvars'])
            # new_asvars_index = self.param.generator.choice(bin, size=size, p=prob)
            # new_asvars = [i for (i, v) in zip(chosen_sol['asvars'], new_asvars_index) if v]
            # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

            # Randomly select a subset of the variables from the chosen solution
            size = int((len(chosen_sol['asvars'])) * prop)
            new_asvars = list(self.param.generator.choice(chosen_sol['asvars'], size=size, replace=False))
            n_asvars = sorted(list(set().union(new_asvars, self.param.ps_asvars)))
            new_asvars = self.remove_redundant_asvars(n_asvars, self.param.trans_asvars, self.param.asvarnames)
            new_sol['asvars'] = new_asvars

            # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            # Randomly select a subset of the variables from the chosen solution
            # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            size = int((len(chosen_sol['isvars'])) * prop)
            new_isvars = list(self.param.generator.choice(chosen_sol['isvars'], size=size, replace=False))
            new_isvars = sorted(list(set().union(new_isvars, self.param.ps_isvars)))
            new_sol['isvars'] = new_isvars

            # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            # Include variables in new solution based on the chosen solution
            # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            new_randvars = {k: v for k, v in chosen_sol['randvars'].items() if k in new_asvars}
            new_sol['randvars'] = new_randvars

            new_bcvars = [var for var in chosen_sol['bcvars']
                            if var in new_asvars and var not in self.param.ps_corvars]
            new_sol['bcvars'] = new_bcvars

            new_corvars = chosen_sol['corvars']
            if new_corvars:
                new_corvars = [var for var in chosen_sol['corvars']
                                if var in new_randvars.keys() and var not in new_bcvars]
            new_sol['corvars'] = new_corvars

            # Take fit_intercept from chosen solution
            new_sol['asc_ind'] = chosen_sol['asc_ind']

            # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            if chosen_sol['class_params_spec'] is not None:
            # {
                class_params_spec = copy.deepcopy(chosen_sol['class_params_spec'])
                for ii, class_params in enumerate(class_params_spec):
                # {
                    class_params_index = self.param.generator.choice(bin, size=len(class_params), p=prob)
                    class_params_spec[ii] = np.array([i for (i, v) in zip(class_params, class_params_index) if v],
                                                     dtype=class_params.dtype)
                # }
                new_sol['class_params_spec'] = class_params_spec
            # }
            # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            if chosen_sol['member_params_spec'] is not None:
            # {
                member_params_spec = copy.deepcopy(chosen_sol['member_params_spec'])
                for ii, member_params in enumerate(member_params_spec):
                # {
                    member_params_index = self.param.generator.choice(bin, size=len(member_params), p=prob)
                    member_params_spec[ii] = np.array([i for (i, v) in zip(member_params, member_params_index) if v],
                                                      dtype=member_params.dtype)
                # }
                new_sol['member_params_spec'] = member_params_spec
            # }
        # }

        return new_sol
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function                                                   '''
    ''' ---------------------------------------------------------- '''
    def remove_non_unique_solutions(self):
    # {
        seen_tuple = set()  # Create an empty set
        new_memory = list()  # Create an empty list
        crit = self.param.criterions[:self.nb_crit]
        for sol in self.memory:
        # {
            sol_tuple = tuple([sol[crit[0]], sol[crit[1]]])
            if sol_tuple not in seen_tuple:
            # {
                seen_tuple.add(sol_tuple)   # Revise what has been seen
                new_memory.append(sol)      # Update list of unique solutions
            # }
        # }
        self.memory = new_memory
    # }

    ''' ------------------------------------------------------------ '''
    ''' Function. Insert solution and filter out non-unique solutions'''
    ''' ------------------------------------------------------------ '''
    def insert_solution(self, solution):
    # {
        self.memory.append(copy.deepcopy(solution))
        self.remove_non_unique_solutions()
        self.memory = self.sort_memory(self.memory)
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Performs the pitch adjustment operation to       '''
    ''' fine-tune a given solution. The process includes adding    '''
    ''' new features or removing existing ones based on a binary   '''
    ''' indicator. The resulting solution is evaluated and inserted'''
    ''' The solutions in memory are then filtered                  '''
    ''' ---------------------------------------------------------- '''
    def pitch_adjustment(self, sol, pitch):
    # {
        # QUERY. IS DEEP COPY REQUIRED?
        adj = copy.deepcopy(sol)  # Adjusted solution

        # pitch adjustment: add/remove as variables
        if self.param.generator.rand() <= pitch: adj = self.perturb_asfeature(sol)
        # pitch adjustment: add|remove is variables
        if self.param.generator.rand() <= pitch: adj = self.perturb_isfeature(adj)
        # pitch adjustment: add|remove random variable
        if self.param.generator.rand() <= pitch: adj = self.perturb_randfeature(adj)

        if self.param.generator.rand() <= pitch: adj = self.change_distribution(adj)

        # pitch adjustment: add|remove bc variables
        if self.param.generator.rand() <= pitch: adj = self.perturb_bcfeature(adj, pitch)
        # Pitch adjustment: add|remove cor variables
        if self.param.generator.rand() <= pitch: adj = self.perturb_corfeature(adj)
        # Pitch adjustment: add|remove class param variables
        if self.param.generator.rand() <= pitch: adj = self.perturb_class_paramfeature(adj)
        # Pitch adjustment: add|remove member param variables
        if self.param.generator.rand() <= pitch: adj = self.perturb_member_paramfeature(adj)

        adj, converged = self.evaluate_solution(adj)
        return adj, converged
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Extracts the best features                       '''
    ''' ---------------------------------------------------------- '''
    def get_best_features(self, memory):
    # {
        soln = self.find_best_sol(memory)

        # Copy necessary values from soln dictionary
        best_asvars = soln['asvars'].copy()
        best_isvars = soln['isvars'].copy()
        best_randvars = soln['randvars'].copy()
        best_bcvars = soln['bcvars'].copy()
        best_corvars = soln['corvars'].copy()
        asc_ind = soln['asc_ind']
        best_class_params_spec = soln['class_params_spec'].copy() if soln['class_params_spec'] is not None else None
        best_member_params_spec = soln['member_params_spec'].copy() if soln['member_params_spec'] is not None else None

        # Return a tuple containing eight things
        return (best_asvars, best_isvars, best_randvars, best_bcvars, best_corvars, asc_ind, best_class_params_spec,
                best_member_params_spec)
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Used in local_search_1 to local_search_10        '''
    ''' ---------------------------------------------------------- '''
    def make_evaluate_insert(self, _asvars, _isvars, _randvars, _bcvars, _corvars, _asc_ind,
                             _class_params_spec, _member_params_spec):
    # {
        solution = Solution(nb_crit=self.nb_crit, asvars=_asvars, isvars=_isvars, randvars=_randvars, bcvars=_bcvars,
                    corvars=_corvars, asc_ind=_asc_ind, class_params_spec=_class_params_spec,
                    member_params_spec=_member_params_spec)

        revised_solution, converged = self.evaluate_solution(solution)
        if converged:
            self.insert_solution(revised_solution)
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Apply local search to improve a solution         '''
    ''' ---------------------------------------------------------- '''
    # Check whether changing a coefficient distribution improves the kpi
    def make_change_1(self, candidate):
    # {
        # Identify the best solution features
        best_asvars, best_isvars, best_randvars, best_bcvars, best_corvars, \
            asc_ind, best_class_params_spec, best_member_params_spec = self.get_best_features(candidate)

        # Make changes
        solution = Solution(nb_crit=self.nb_crit, asvars=best_asvars, isvars=best_isvars, randvars=best_randvars, bcvars=best_bcvars,
                            corvars=best_corvars, asc_ind=asc_ind, class_params_spec=best_class_params_spec,
                            member_params_spec=best_member_params_spec)

        revised_solution = self.change_distribution(solution)  # Make a change

        # Revise the following dictionary and lists
        best_randvars = {key: val for key, val in best_randvars.items() if key in best_asvars and val != 'f'}
        best_bcvars = [var for var in best_bcvars if var in best_asvars and var not in self.param.ps_corvars]
        best_corvars = [var for var in best_randvars.keys() if var not in best_bcvars]

        # Make a solution, evaluate it, and insert if converged
        self.make_evaluate_insert(best_asvars, best_isvars, best_randvars, best_bcvars, best_corvars, asc_ind,
                            best_class_params_spec, best_member_params_spec)
    # }

    # Check if having a full covariance matrix leads to an improved BIC
    def make_change_2(self, candidate):
    # {
        # Identify the best solution features
        best_asvars, best_isvars, best_randvars, best_bcvars, best_corvars, \
            asc_ind, best_class_params_spec, best_member_params_spec = self.get_best_features(candidate)

        # Make changes
        best_bcvars = [var for var in best_asvars if var in self.param.ps_bcvars]
        if self.param.ps_cor is None or self.param.ps_cor:
            best_corvars = [var for var in best_randvars if var not in best_bcvars]
        else:
            best_corvars.clear()

        # Make a solution, evaluate it, and insert if converged
        self.make_evaluate_insert(best_asvars, best_isvars, best_randvars, best_bcvars, best_corvars, asc_ind,
                          best_class_params_spec, best_member_params_spec)
    # }

    # Check if having all the variables transformed leads to an improvement in BIC
    def make_change_3(self, candidate):
    # {
        # Identify the best solution features
        best_asvars, best_isvars, best_randvars, best_bcvars, best_corvars, \
            asc_ind, best_class_params_spec, best_member_params_spec = self.get_best_features(candidate)

        # Make changes
        if self.param.ps_bctrans is None or self.param.ps_bctrans:
            best_bcvars = [var for var in best_asvars if var not in self.param.ps_corvars]
        else:
            best_bcvars.clear()

        best_corvars = [var for var in best_randvars.keys() if var not in best_bcvars]

        # Make a solution, evaluate it, and insert if converged
        self.make_evaluate_insert(best_asvars, best_isvars, best_randvars, best_bcvars, best_corvars, asc_ind,
                          best_class_params_spec, best_member_params_spec)
    # }

    def make_change_4(self, candidate):
    # {
        # Identify the best solution features
        best_asvars, best_isvars, best_randvars, best_bcvars, best_corvars, \
            asc_ind, best_class_params_spec, best_member_params_spec = self.get_best_features(candidate)

        if len(best_asvars) < len(self.param.asvarnames):
        # {
            solution = Solution(nb_crit=self.nb_crit, asvars=best_asvars, isvars=best_isvars,
                                randvars=best_randvars, bcvars=best_bcvars,
                                corvars=best_corvars, asc_ind=asc_ind,
                                class_params_spec=best_class_params_spec,
                                member_params_spec=best_member_params_spec)

            solution = self.add_asfeature(solution)
            revised_solution, converged = self.evaluate_solution(solution)
            if converged: self.insert_solution(revised_solution)
        # }
    # }


    def make_change_5(self, candidate):
    # {
        # Identify the best solution features
        best_asvars, best_isvars, best_randvars, best_bcvars, best_corvars, \
            asc_ind, best_class_params_spec, best_member_params_spec = self.get_best_features(candidate)

        solution = Solution(nb_crit=self.nb_crit, asvars=best_asvars, isvars=best_isvars,
                            randvars=best_randvars, bcvars=best_bcvars, corvars=best_corvars, asc_ind=asc_ind,
                            class_params_spec=best_class_params_spec, member_params_spec=best_member_params_spec)

        solution = self.add_isfeature(solution)
        revised_solution, converged = self.evaluate_solution(solution)
        if converged: self.insert_solution(revised_solution)
    # }

    def make_change_6(self, candidate):
    # {
        # Identify the best solution features
        best_asvars, best_isvars, best_randvars, best_bcvars, best_corvars, \
            asc_ind, best_class_params_spec, best_member_params_spec = self.get_best_features(candidate)

        solution = Solution(nb_crit=self.nb_crit, asvars=best_asvars, isvars=best_isvars, randvars=best_randvars,
                            bcvars=best_bcvars, corvars=best_corvars, asc_ind=asc_ind,
                            class_params_spec=best_class_params_spec, member_params_spec=best_member_params_spec)

        if self.param.avail_bcvars: # {
            solution = self.perturb_bcfeature(solution, self.pitch)
            revised_solution, converged = self.evaluate_solution(solution)
            if converged: self.insert_solution(revised_solution)
        # }
    # }

    def make_change_7(self, candidate):
    # {
        # Identify the best solution features
        best_asvars, best_isvars, best_randvars, best_bcvars, best_corvars, \
            asc_ind, best_class_params_spec, best_member_params_spec = self.get_best_features(candidate)

        solution = Solution(nb_crit=self.nb_crit, asvars=best_asvars, isvars=best_isvars, randvars=best_randvars, bcvars=best_bcvars,
                            corvars=best_corvars, asc_ind=asc_ind, class_params_spec=best_class_params_spec,
                            member_params_spec=best_member_params_spec)

        if self.param.avail_rvars:
        # {
            solution = self.perturb_corfeature(solution)
            revised_solution, converged = self.evaluate_solution(solution)
            if converged:
                self.insert_solution(revised_solution)

            revised_solution = self.perturb_randfeature(solution)
            revised_solution, converged = self.evaluate_solution(revised_solution)
            if converged:
                self.insert_solution(revised_solution)
        # }
    # }

    # Check if changing coefficient distributions improves the solution BIC
    def make_change_8(self, candidate):
    # {
        # Identify the best solution features
        best_asvars, best_isvars, best_randvars, best_bcvars, best_corvars, \
            asc_ind, best_class_params_spec, best_member_params_spec = self.get_best_features(candidate)

        for var in best_randvars: # {
            if var not in self.param.ps_randvars:
                rm_dist = [distr for distr in self.param.distr if distr != best_randvars[var]]
                best_randvars[var] = self.param.generator.choice(rm_dist)
        # }
        best_randvars = {key: val for key, val in best_randvars.items() if key in best_asvars and val != 'f'}
        best_bcvars = [var for var in best_bcvars if var in best_asvars and var not in self.param.ps_corvars]

        if self.param.ps_cor is None or self.param.ps_cor:
            best_corvars = [var for var in best_randvars.keys() if var not in best_bcvars]
        else:
            best_corvars = []

        if len(best_corvars) < 2:
            best_corvars = []

        solution = Solution(nb_crit=self.nb_crit, asvars=best_asvars, isvars=best_isvars,
                            randvars=best_randvars, bcvars=best_bcvars, corvars=best_corvars,
                            asc_ind=asc_ind, class_params_spec=best_class_params_spec,
                            member_params_spec=best_member_params_spec)
        revised_solution, converged = self.evaluate_solution(solution)

        if converged:
            self.insert_solution(revised_solution)
    # }

    def make_change_9(self, candidate):
    # {
        # Identify the best solution features
        best_asvars, best_isvars, best_randvars, best_bcvars, best_corvars, \
            asc_ind, best_class_params_spec, best_member_params_spec = self.get_best_features(candidate)


        solution = Solution(nb_crit=self.nb_crit, asvars=best_asvars, isvars=best_isvars,
                randvars=best_randvars, bcvars=best_bcvars, corvars=best_corvars, asc_ind=asc_ind,
                class_params_spec=best_class_params_spec, member_params_spec=best_member_params_spec)

        if solution['class_params_spec'] is not None:
        # {
            revised_solution = self.perturb_class_paramfeature(solution)
            revised_solution, converged = self.evaluate_solution(revised_solution)
            if converged:
                self.insert_solution(revised_solution)
        # }
    # }

    def make_change_10(self, candidate):
    # {
        # Identify the best solution features
        best_asvars, best_isvars, best_randvars, best_bcvars, best_corvars, \
            asc_ind, best_class_params_spec, best_member_params_spec = self.get_best_features(candidate)

        solution = Solution(nb_crit=self.nb_crit, asvars=best_asvars, isvars=best_isvars,
                    randvars=best_randvars, bcvars=best_bcvars, corvars=best_corvars, asc_ind=asc_ind,
                    class_params_spec=best_class_params_spec, member_params_spec=best_member_params_spec)

        if solution['member_params_spec'] is not None:
        # {
            revised_solution = self.perturb_member_paramfeature(solution)
            revised_solution, converged = self.evaluate_solution(revised_solution)
            if converged: self.insert_solution(revised_solution)
        # }
    # }

    def local_search(self):
    # {
        # Identify candidate solutions
        candidate = [sol for sol in self.memory if abs(sol.obj[0]) < BOUND]

        # TODO
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function                                                   '''
    ''' ---------------------------------------------------------- '''
    def log_convergence(self, memory):
    # {
        crit = self.param.criterions[:self.nb_crit]

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Filter out poor solutions:
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        filtered_memory = [sol for sol in memory if abs(sol[crit[0]]) < BOUND and abs(sol[crit[1]]) < BOUND]
        # OR,
        # filtered_memory = [sol for sol in memory if abs(sol.obj[0]) < BOUND and abs(sol.obj[1]) < BOUND]

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Sort the new list of solutions by 'sol_num':
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        filtered_memory = sorted(filtered_memory, key=lambda sol: sol['sol_num'])

        # ~~~~~~~~~~~~~~~~~~~~~~~~~
        # Record the best obj val
        # ~~~~~~~~~~~~~~~~~~~~~~~~~
        best_val = self.get_best_val(self.param.criterions, filtered_memory)
        all_val = self.get_all_val(self.param.criterions, filtered_memory)
        for i in range(self.nb_crit):
            logger.debug(f"Best points (obj {i}): {best_val[i]}") # Log the ith value

        return all_val, best_val
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function  Conduct harmony memory consideration             '''
    ''' pitch adjustment, and local search                         '''
    ''' This function tracks the progress of the optimization      '''
    ''' process by recording the score of the best and current     '''
    ''' solutions at each iteration                                '''
    ''' ---------------------------------------------------------- '''
    def improvise(self):
    # {
        best, current = [], []
        for iter in range(self.maxiter):
        # {
            # Compute consideration rate and pitch value
            # This code introduces oscillations (a.k.a., variations) based on the iteration number.
            # The result is scaled by the sine function only when its value is non-negative.
            sine_iter = max(0, np.sign(math.sin(iter)))
            self.harm_rate = (self.min_harm + ((self.max_harm - self.min_harm) / self.maxiter) * iter) * sine_iter
            self.pitch = (self.min_pitch + ((self.max_pitch - self.min_pitch) / self.maxiter) * iter) * sine_iter

            new_sol = self.build_solution(self.memory, self.harm_rate) # Create a single new solution and perform an adjustment
            curr_sol, converged = self.pitch_adjustment(new_sol, self.pitch)   # Perform additional perturbations
            if converged:
            # {
                self.insert_solution(curr_sol)

                #if iter > int(self.prop_local * self.maxiter):
                # {
                    # Run local search
                    #best_sol = self.memory[0]
                    #best.append(best_sol.obj[0])
                    #current.append(curr_sol.obj[0])
                # }
            # }
        # }

        all_val, obj_val = self.log_convergence(self.memory)
        if self.generate_plots:
            self.plot_results(self.memory, all_val, obj_val)
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.  Discard non-convergent solutions                '''
    ''' ---------------------------------------------------------- '''
    def screen_solutions(self, solutions):
    # {
        feasible_solutions = []
        if solutions is not None:
        # {
            for sol in solutions:
            # {
                new_sol = copy.deepcopy(sol)
                new_sol = self.increase_sol_by_one_class(new_sol)
                new_sol.pop('class_num')    # Remove 'class_num'
                new_sol, converged = self.evaluate_solution(new_sol)
                if converged:
                    feasible_solutions.append(new_sol)
            # }
        # }
        return feasible_solutions
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Code used in "self.run_search"                   '''
    ''' ---------------------------------------------------------- '''
    def extract_parameter(self):
    # {
        avail, avail_latent = self.param.avail, self.param.avail_latent
        weights = self.param.weights
        alt_var = self.param.alt_var
        choice_id = self.param.choice_id
        ind_id = self.param.ind_id

        if self.nb_crit > 1:
        # {
            if self.param.avail is not None:
                avail = np.row_stack((self.param.avail, self.param.test_avail))

            if self.param.avail_latent is not None:
            # {
                avail_latent = make_list(None, self.param.num_classes)  # i.e., [None] * self.param.num_classes
                for ii, avail_latent_ii in enumerate(self.param.avail_latent):
                    # {
                    if avail_latent_ii is not None:
                        avail_latent[ii] = np.row_stack((avail_latent_ii, self.param.test_avail_latent[ii]))
                # }
            # }

            if self.param.weights is not None:
                weights = np.concatenate((self.param.weights, self.param.test_weight_var))

            if self.param.alt_var is not None:
                alt_var = np.concatenate((self.param.alt_var, self.param.test_alt_var))

            if self.param.choice_id is not None:
                choice_id = np.concatenate((self.param.choice_id, self.param.test_choice_id))

            if self.param.ind_id is not None:
                ind_id = np.concatenate((self.param.ind_id, self.param.test_ind_id))
        # }
        return  avail, weights, alt_var, choice_id, ind_id, avail_latent
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def extract_from_sol(self, sol):
    # {
        asvarnames, isvarnames, randvars, bcvars, corvars, intercept, \
            class_params_spec, member_params_spec = self.get_components(sol)

        # Revise varnames
        if self.param.latent_class:
            varnames = np.concatenate(class_params_spec + member_params_spec + [isvarnames])
            varnames = np.unique(varnames)
        else:
            varnames = asvarnames + isvarnames

        # Delete '_inter' bug fix
        if '_inter' in varnames:
            varnames = np.delete(varnames, np.argwhere(varnames == '_inter'))

        return varnames, asvarnames, isvarnames, randvars, bcvars, corvars, \
            intercept, class_params_spec, member_params_spec
    # }

    # call if self.param.multi_objective?
    def test_best_solution(self, best_sol):
    # {
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Sort memory and extract features
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        best_varnames, best_asvarnames, best_isvarnames, best_randvars, best_bcvars, best_corvars, \
            best_intercept, best_class_params_spec, best_member_params_spec = self.extract_from_sol(best_sol)

        avail_all, weights_all, alt_var_all, choice_id_all, ind_id_all, avail_latent_all = self.extract_parameter()

        df_all = pd.concat([self.param.df, self.param.df_test], ignore_index=True)
        y = self.param.choices + self.param.test_choices
        X = df_all[best_varnames]

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Define appropriate model and fit coefficients
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        if bool(best_randvars):
        # {
            if self.param.latent_class:
                """
                Note template: search::fit_lcmm(X, y, varnames, isvars, class_params_spec, member_params_spec, num_classes,
                    alts, ids, panels, bcvars, randvars, corvars, maxiter, gtol, avail, weights)
                """
                model = self.fit_lcmm(X, y, best_varnames, best_isvarnames, best_class_params_spec,
                    best_member_params_spec, self.param.num_classes, alt_var_all, choice_id_all,
                    ind_id_all, best_bcvars, best_randvars, best_corvars,
                    self.param.maxiter, self.param.gtol, avail_all, weights_all)
            else:
                """
                Note template: search::fit_mxl(X, y, varnames, alts, isvars, transvars, ids, panels, randvars, corvars,
                    fit_intercept, n_draws, weights, avail, base_alt,  maxiter, seed, ftol, gtol, save_fitted_params)
                """
                model = self.fit_mxl(X, y, best_varnames, alt_var_all, best_isvarnames, best_bcvars,
                    choice_id_all, ind_id_all, best_randvars, best_corvars, best_intercept,
                    self.param.n_draws, None, None, None, 2000, None, 1e-6, 1e-6, False)
            # }
        # }
        else:
        # {
            if self.param.latent_class:
                """ 
                Note template: search::fit_lcm(X, y, varnames, class_params_spec, member_params_spec, num_classes, ids,
                transvars, maxiter, gtol, gtol_membership_func, avail, avail_latent, intercept_opts, weights, seed,
                alts, ftol_lccm, base_alt)
                """
                seed = self.param.generator.randint(2 ** 31 - 1)
                model = self.fit_lcm(X, y, best_varnames, best_class_params_spec, best_member_params_spec,
                    self.param.num_classes, choice_id_all, best_bcvars, self.param.maxiter,
                    self.param.gtol, self.param.gtol_membership_func, avail_all, avail_latent_all,
                    self.param.intercept_opts, weights_all, seed, None, 1e-6, None)

            else:
                """
                Note template: search::fit_mnl(X, y, varnames, isvars, alts, ids, transvars, fit_intercept,
                    weights, avail, base_alt, maxiter, ftol, gtol, seed)
                """
                model = self.fit_mnl(X, y, best_varnames, best_isvarnames, alt_var_all,
                    choice_id_all, best_bcvars, best_intercept, None, None, None, 2000, 1e-6, 1e-6, None)
        # }

        report_model_statistics(self.results_file, model)  # Output the model statistics
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def log_solutions(self, solutions):
    # {
        if self.nb_crit == 1:
        # {
            all_solutions = sorted(solutions, key=lambda sol: sol.obj[0])
            best_sols = all_solutions[:self.max_mem]
            best_sol = best_sols[0]
            logger.info("Model with best score had {} classes".format(self.max_classes))
            logger.info("Best solution")
            for k, v in best_sol.items():
                logger.info(f"{k}: {v}")
        # }
        else:
        # {
            fronts = self.get_fronts(solutions)
            pareto = self.get_pareto(fronts, solutions)
            all_solutions = self.non_dominant_sorting(solutions)
            logger.info("Models in Pareto front had at most {} classes".format(self.max_classes))
            logger.info("Best models in Pareto front")
            for i, sol in enumerate(pareto):
            # {
                logger.info(f'Best solution - {i}')
                for k, v in sol.items():
                    logger.info(f"{k}: {v}")
            # }
            best_sol = all_solutions[0]
            logger.info(f"Best solution with {best_sol['class_num']} classes")
            for k, v in best_sol.items():
                logger.info(f"{k}: {v}")
        # }
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def run_search(self, existing_sols=None):
    # {
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Combine solutions into one list
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        existing_memory = self.screen_solutions(existing_sols)  # Screen out non convergent solutions
        generated_memory = self.initialize_memory(self.max_mem)  # Generate some solutions

        # OPTIONAL: CREATE max(max_mem - existing, 0) NEW SOLUTIONS?

        init_memory = generated_memory + existing_memory    # Aggregate solution lists
        unique_memory = get_unique(init_memory, 0)  # Remove duplicate solutions if present
        for sol in unique_memory:
            sol.data['is_initial_sol'] = True  # Set solution status

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Sort memory by first objective or by Pareto ranking and retain specified number
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        memory_sorted = self.sort_memory(unique_memory)
        memory = memory_sorted[:self.max_mem]
        self.memory = memory.copy()

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Generate new solutions by combining elements from existing ones
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        self.improvise()

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Copy, Sort & Test
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        memory = self.memory.copy()
        improved_memory = self.sort_memory(memory)
        best_sol = improved_memory[0]
        self.test_best_solution(best_sol)

        # ~~~~~~~~~~~~~~~~
        # Perform logging
        # ~~~~~~~~~~~~~~~~
        logger.info("Improved harmony: {}".format(improved_memory))
        logger.info("Search ended at: {}".format(str(time.ctime())))

        return improved_memory
    # }


    def run_search_latent(self, override=False):
    # {
        prev, best_model_idx = infinity, 0
        all_solutions, solutions = [], []

        for q in range(self.min_classes, self.max_classes + 1):
        # {
            self.param.num_classes = q
            self.param.latent_class = False if q == 1 else True
            solutions = self.run_search(existing_sols=all_solutions)

            # This code iterates over each dictionary sol in the solutions list and updates
            # the value associated with the key 'class_num' to q. The use of a list
            # comprehension avoids the need for an explicit loop.
            [sol.update({'class_num': q}) for sol in solutions]

            # Aggregate solutions
            all_solutions = all_solutions + solutions

            if self.param.nb_crit > 1:
            # {
                all_solutions = self.non_dominant_sorting(all_solutions)
                fronts = self.get_fronts(all_solutions)
                pareto = self.get_pareto(fronts, all_solutions)
                self.pareto_front = pareto

                # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                # Check if a solution with q classes is in the Pareto front
                # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                pareto_class_nums = [sol['class_num'] for sol in pareto]
                stop_run = max(pareto_class_nums) != q
                if stop_run and not override:
                    logger.info(f"Stopping search at {q} classes")
                    break  # Exit the loop immediately

                best_model_idx += 1
            # }
            else:
            # {
                all_solutions = sorted(solutions, key=lambda sol: sol.obj[0])
                best_solution = all_solutions[0]  # assume already sorted
                if best_solution.obj[0] < prev or override:
                # {
                    best_model_idx += 1
                    prev = best_solution.obj[0]
                # }
                else:  # {
                    break  # Exit the loop immediately
                # }
            # }
        # }
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        self.log_solutions(all_solutions)
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        best_val, all_val, all_val_classes = self.post_process(all_solutions)

        if self.generate_plots:
            self.plot_results_latent(all_solutions, best_val, all_val, all_val_classes)

        return all_solutions
    # }


    ''' ---------------------------------------------------------- '''
    ''' Function                                                   '''
    ''' ---------------------------------------------------------- '''
    def plot_multi(self, solutions, all_val):
    # {
        crit = self.param.criterions[:self.nb_crit]
        fig, ax = plt.subplots()

        # ~~~~~~~~~~~
        # LINE 1
        # ~~~~~~~~~~~
        scaled_val = np.log(all_val[1]) if crit[1] == 'MAE' else all_val[1]
        line_1 = ax.scatter(all_val[0], scaled_val, label="All solutions", marker='o')

        # ~~~~~~~~~~~
        # LINE 2
        # ~~~~~~~~~~~
        init_solns = [sol for sol in solutions if abs(sol.obj[0]) < BOUND]
        init_0 = [sol[crit[0]] for sol in init_solns]
        init_1 = [sol[crit[1]] for sol in init_solns]
        line_2 = ax.scatter(init_0, init_1, label="Initial solutions", marker='x')

        # ~~~~~~~~~~~~~~~~~~~~
        # PARETO CALCULATIONS
        # ~~~~~~~~~~~~~~~~~~~~
        fronts = self.get_fronts(solutions)
        pareto = self.get_pareto(fronts, solutions)
        self.pareto_front = [sol for sol in pareto if abs(sol.obj[0]) < BOUND]  # Store filtered
        pareto_0 = np.array([sol.obj[0] for sol in pareto])
        pareto_1 = np.array([sol.obj[1] for sol in pareto])
        pareto_1 = np.log(pareto_1) if crit[1] == 'MAE' else pareto_1

        # ~~~~~~~~~~~
        # LINE 3
        # ~~~~~~~~~~~
        line_3 = ax.scatter(pareto_0, pareto_1, label="Pareto Front", marker='o')

        # ~~~~~~~~~~~
        # LINE 4
        # ~~~~~~~~~~~
        # Determine the indices that would sort the pareto_0 array in ascending order.
        # These indices represent the positions of elements in the original array.
        pareto_idx = np.argsort(pareto_0)
        line_4 = ax.plot(pareto_0[pareto_idx], pareto_1[pareto_idx], color="r", label="Pareto Front")

        # ~~~~~~~~~~~
        # ALL LINES
        # ~~~~~~~~~~~
        all_lines = (line_1, line_2, line_4[0])

        labels = [line.get_label() for line in all_lines]
        log_str = 'log' if crit[1] == 'MAE' else ''
        ax.set_xlabel(f"{crit[0]} - Training dataset")
        ax.set_ylabel(f"{crit[1]} - Testing dataset")
        lgd = ax.legend(all_lines, labels, loc='upper right', bbox_to_anchor=(0.5, -0.1))
        current_time = datetime.datetime.now().strftime("%d%m%Y-%H%M%S")
        latent_info = "_" + str(self.param.num_classes) + "_classes_" if (self.param.num_classes > 1) else "_"
        plot_filename = self.code_name + "_" + latent_info + current_time + "_MOOF.png"
        plt.savefig(plot_filename, bbox_extra_artists=(lgd,), bbox_inches='tight')
    # }

    def plot_multi_latent(self, solutions, all_val_classes):
    #
        crit = self.param.criterions[:self.nb_crit]

        fronts = self.get_fronts(solutions)
        pareto = self.get_pareto(fronts, solutions)
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        fig, ax = plt.subplots()
        lns_all = []
        for q in range(self.min_classes,self.max_classes):
        # {
            class_label = "All solutions - " + str(q) + " classes"
            if q == 1:
                class_label = "All solutions - " + str(q) + " class"
            lns = ax.scatter(all_val_classes[0][q], all_val_classes[1][q], label=class_label, marker='o')
            lns_all.append(lns)
        # }
        # ~~~~~~~~~~~
        # LINE 2
        # ~~~~~~~~~~~
        init_val = [[] for _ in range(self.nb_crit)]
        init_sols = [sol for sol in solutions if sol.obj[0] < BOUND and sol['is_initial_sol']]
        init_val[0] = [sol.obj[0] for sol in init_sols]
        if crit[0] == 'MAE': init_val[1] = np.log(init_val[1])
        init_val[1] = [sol.obj[1] for sol in init_sols]
        if crit[1] == 'MAE': init_val[1] = np.log(init_val[1])
        lns2 = ax.scatter(init_val[0], init_val[1], label="Initial solutions", marker='x', color='black')

        # ~~~~~~~~~~~
        # LINE 4
        # ~~~~~~~~~~~
        pareto = [pareto for _, pareto in enumerate(pareto) if np.abs(pareto.obj[0]) < BOUND]
        logger.info('Final Pareto: {}'.format(str(pareto)))
        pareto_0 = np.array([par.obj[0] for par in pareto])
        pareto_1 = np.array([par.obj[1] for par in pareto])
        log_str = ''
        if crit[1] == 'MAE':
            pareto_1 = np.log(pareto_1)
            log_str = 'log'

        pareto_idx = np.argsort(pareto_0)
        lns4 = ax.plot(pareto_0[pareto_idx], pareto_1[pareto_idx], color="r", label="Pareto Front")

        # ~~~~~~~~~~~
        # ALL LINES
        # ~~~~~~~~~~~
        lns = (*lns_all, lns2, lns4[0])

        labs = [l_pot.get_label() for l_pot in lns]
        ax.set_xlabel(f"{crit[0]} - Training dataset")
        ax.set_ylabel(f"{log_str} {crit[1]} - Testing dataset")
        lgd = ax.legend(lns, labs, loc='upper center', bbox_to_anchor=(0.5, -0.1))
        current_time = datetime.datetime.now().strftime("%d%m%Y-%H%M%S")
        plot_filename = self.code_name + "_" + current_time + "_MOOF.png"
        plt.savefig(plot_filename, bbox_extra_artists=(lgd,), bbox_inches='tight')
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function                                                   '''
    ''' ---------------------------------------------------------- '''
    def plot_single(self, all_val, best_val):
    # {
        fig, ax1 = plt.subplots()
        ax2 = ax1.twinx()
        ax1.xaxis.get_major_locator().set_params(integer=True)

        crit = self.param.criterions[:self.nb_crit]
        label ="Solution estimated at current iteration (" + crit + ")"
        line_1 = ax1.plot(np.arange(len(all_val[0])), all_val[0], label=label)

        label = "Best solution in memory at current iteration (" + crit + ")"
        line_2 = ax1.plot(np.arange(len(best_val[0])), best_val[0], label=label, linestyle="dotted")

        label = "In-sample LL of best solution in memory at current iteration"
        line_3 = ax2.plot(np.arange(len(best_val[1])), best_val[1], label=label, linestyle="dashed")

        all_lines = line_1 + line_2 + line_3

        labels = [line.get_label() for line in all_lines]
        handles, _ = ax1.get_legend_handles_labels()
        lgd = ax1.legend(all_lines, labels, loc='upper center', bbox_to_anchor=(0.5, -0.1))
        ax1.set_xlabel("Iterations")
        ax1.set_ylabel(crit[0])
        ax2.set_ylabel(crit[1])
        current_time = datetime.datetime.now().strftime("%d%m%Y-%H%M%S")
        latent_info = "_" + str(self.param.num_classes) + "_classes_" if (self.param.num_classes > 1) else "_"
        plot_filename = self.code_name + "_" + latent_info + current_time + "_SOOF.png"
        plt.savefig(plot_filename, bbox_extra_artists=(lgd,), bbox_inches='tight')
    # }

    def plot_single_latent(self, best_val, all_val, all_val_classes):
    # {
        fig, ax1 = plt.subplots()
        ax2 = ax1.twinx()
        ax1.xaxis.get_major_locator().set_params(integer=True)

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        counter = 0
        max_0 = np.max(all_val[0])
        for q in range(self.min_classes, self.max_classes):
        # {
            num_sols_in_class = len(all_val_classes[0][q])
            ax1.axvline(x=counter, color='r', linestyle='--')
            if q == 1: line_text = '1 class'
            else:  line_text = str(q) + ' classes'
            ax1.text(counter, max_0, line_text)
            counter += num_sols_in_class
        # }
        all_val[0] = np.concatenate(np.array(all_val_classes[0]))

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        crit = self.param.criterions[:self.nb_crit]
        label ="Solution estimated at current iteration (" + crit + ")"
        line_1 = ax1.plot(np.arange(len(all_val[0])), all_val[0], label=label)

        label = "Best solution in memory at current iteration (" + crit + ")"
        line_2 = ax1.plot(np.arange(len(best_val[0])), best_val[0], label=label, linestyle="dotted")

        label = "In-sample LL of best solution in memory at current iteration"
        line_3 = ax2.plot(np.arange(len(best_val[1])), best_val[1], label=label, linestyle="dashed")

        all_lines = line_1 + line_2 + line_3

        labels = [line.get_label() for line in all_lines]
        handles, _ = ax1.get_legend_handles_labels()
        lgd = ax1.legend(all_lines, labels, loc='upper center', bbox_to_anchor=(0.5, -0.1))
        ax1.set_xlabel("Iterations")
        ax1.set_ylabel(crit[0])
        ax2.set_ylabel(crit[1])
        current_time = datetime.datetime.now().strftime("%d%m%Y-%H%M%S")
        latent_info = "_" + str(self.param.num_classes) + "_classes_" if (self.param.num_classes > 1) else "_"
        plot_filename = self.code_name + "_" + latent_info + current_time + "_SOOF.png"
        plt.savefig(plot_filename, bbox_extra_artists=(lgd,), bbox_inches='tight')
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function                                                   '''
    ''' ---------------------------------------------------------- '''
    def post_process(self, solutions):
    # {
        valid_solutions = [sol for sol in solutions if sol.obj[0] < BOUND]
        valid_solutions = sorted(valid_solutions, key=lambda sol: sol['sol_num'])

        all_val_classes, all_val = [[] for _ in range(self.nb_crit)]
        for q in range(self.min_classes, self.max_classes + 1):
        # {
            for i in range(self.nb_crit):
            # {
                all_val[i] = [sol.obj[i] for sol in valid_solutions if sol['class_num'] == q]
                crit = self.param.crit(i)
                if crit == 'MAE': all_val[i] = np.log(all_val[i])
                all_val_classes[i].append(all_val[i])
            # }
        # }

        best_val = [[] for _ in range(self.nb_crit)]
        for i in range(self.nb_crit):
            best_val[i] = self.get_best_val(self.param.criterions, all_val_classes[i])

        return best_val, all_val, all_val_classes
    # }
    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def plot_results(self, solutions, best_val, all_val):
    # {
        if self.nb_crit == 1:
            self.plot_single(all_val[0], best_val)
        else:
            self.plot_multi(solutions, all_val)
    # }

    def plot_results_latent(self, solutions, best_val, all_val, all_val_classes):
    # {
        if self.nb_crit == 1:
            self.plot_single_latent(best_val, all_val, all_val_classes)
        else:
            self.plot_multi_latent(solutions, all_val_classes)
    # }
# }




"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
IMPLEMENTATION: SIMULATED ANNEALING
"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
#from scipy.special import kwargs

"""
BACKGROUND - SIMULATED ANNEALING:

Simulated annealing is a probabilistic optimization algorithm inspired by the annealing process in metallurgy.
In metallurgy, annealing is a heat treatment process where a material (typically metal) is heated to a 
certain temperature and then cooled slowly to remove defects and reduce hardness. Similarly, in simulated annealing, 
the algorithm tries to find the global optimum of a function by iteratively exploring the solution space while 
gradually reducing the probability of accepting worse solutions as it progresses.

Here's how simulated annealing typically works:

1. Initialization: Start with an initial solution to the optimization problem. This could be a randomly generated 
solution or some other method depending on the problem.

2. Temperature Schedule: Simulated annealing uses a temperature parameter that controls the probability of accepting 
worse solutions as the algorithm progresses. Initially, the temperature is set to a high value, allowing the 
algorithm to explore a wide range of solutions. As the algorithm progresses, the temperature is gradually 
decreased according to a predefined cooling schedule.

3. Neighbor Generation: At each iteration, a neighboring solution to the current solution is generated. The neighbor
 could be obtained by making a small change to the current solution, such as flipping a bit in a binary 
 representation or perturbing the current solution in some other way.

4. Acceptance Criterion: The algorithm evaluates the quality of the neighboring solution using an objective 
function (fitness function). If the neighboring solution is better than the current solution, 
it is always accepted. If the neighboring solution is worse, it may still be accepted with a certain 
probability determined by the Metropolis criterion.

Iteration: Repeat steps 3 and 4 for a certain number of iterations or until a stopping criterion is met 
(e.g., reaching a maximum number of iterations, convergence criteria).

Cooling Schedule: The temperature is reduced gradually according to a predefined cooling schedule. 
Common cooling schedules include linear, exponential, or logarithmic cooling.

Simulated annealing is a versatile optimization algorithm that can be applied to various optimization 
problems, including combinatorial optimization, continuous optimization, and machine learning tasks. 
It's particularly useful for problems where the objective function is non-convex or has multiple 
local optima, as it allows the algorithm to escape local optima and explore the solution space more thoroughly.
"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""

''' ---------------------------------------------------------- '''
''' LIBRARIES                                                  '''
''' ---------------------------------------------------------- '''
from search import*
import copy
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
import threading
from typing import Callable, Tuple
from datetime import datetime
import random
import string
''' ---------------------------------------------------------- '''

overall_best_solution = None  # PARSA: Reference to best solution
lock = threading.Lock()  # PARSA: Mutex - synchronization primitive


def generate_random_run_name(prefix="run"):
    # Generate a random string of 6 characters
    random_suffix = ''.join(random.choices(string.ascii_letters + string.digits, k=4))

    # Get the current timestamp
    timestamp = datetime.now().strftime("%m%d_%H%M")

    # Combine the elements to form the run name
    run_name = f"{prefix}_{timestamp}_{random_suffix}"
    return run_name

''' ---------------------------------------------------------- '''
''' CLASS FOR SIMULATED ANNEALING                              '''
''' ---------------------------------------------------------- '''
class SA(Search):
# {
    """ Docstring """

    ''' ---------------------------------------------------------- '''
    ''' Function. Constructor                                      '''
    ''' ---------------------------------------------------------- '''
    def __init__(self, param:Parameters, init_sol, ctrl, idnum=generate_random_run_name(), **kwargs):
    # {
        super().__init__(param, idnum, **kwargs)     # Call base class constructor

        tI, tF, max_temp_steps, max_iter = ctrl  # Extract form 'ctrl'


        self.start_time = time.time()
        # Set parameters:
        self.max_time = kwargs.get('max_time', 3600*12)    # Maximum Allowable Run Time (Terminate
        self.max_total_iter = kwargs.get('max_total_iter', 10000)
        self.tI = tI                # Starting temperature
        self.tF = tF                # Final temperature
        self.max_temp_steps = max_temp_steps    # Maximum number of temperature steps
        self.max_iter = max_iter    # Maximum number of iterations at each temperature step
        self.max_no_impr = 3        # Max number of steps permitted without improvements
        self.terminate = False      # Termination flag
        self.rate = np.exp((1.0 / (self.max_temp_steps-1)) * np.log(self.tF/self.tI)) # Temperature reduction rate

        # Note: tF = tI * power(rate, max_temp_steps-1)
        # Note: Subtract one because the first step at t=tI must be included as a step

        self.no_impr = 0            # Current number of iterations without improvement
        self.step = 0               # Current temperature step
        self.t = 0                  # Current temperature
        self.current_sol = init_sol # Current solution
        self.best_sol = None        # Best solution
        self.archive = []           # Archive of solutions
        self.start = None           # Start time
        self.accepted, self.not_accepted = 0, 0 # Counters
        self.comm_int = 1           # Communication interval for PARSA
        self.idnum = idnum

        # Outputting results and convergence information
        self.open_files()

        # Define a member function for the acceptance function
        Args = Tuple[np.ndarray, np.ndarray]
        AcceptanceFn = Callable[[Args], bool]
        self.accept_change: AcceptanceFn = self.accept_change_single \
            if self.nb_crit == 1 else self.accept_change_multi

        # Define a member function for the perturbation function
        PerturbFn = Callable[[], None]
        self.perturb_function: PerturbFn = self.perturb_single if self.nb_crit == 1 else self.perturb_multi

    # }

    def get_run_time(self):
        '''Gets the current run_time in seconds'''
        end_time = time.time()
        return end_time - self.start_time

    def open_files(self):
    # {
        str_idnum = str(self.idnum)
        self.results_file = open("siman_results[" + str_idnum + "].txt", "w")
        self.progress_file = open("siman_progress[" + str_idnum + "].txt", "w")
        self.archive_file = open("siman_archive[" + str_idnum + "].txt", "w")
        self.debug_file = open("siman_pert[" + str_idnum + "].txt", "w")
        self.best_file = open("siman_best[" + str_idnum + "].txt", "w")
    # }

    def close_files(self):
    # {
        self.results_file.close()
        self.progress_file.close()
        self.archive_file.close()
        self.debug_file.close()
        self.best_file.close()
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def curr_score(self, i):
        return self.current_sol.obj(i)

    def best_score(self, i):
        return self.best_sol.obj(i)

    def return_best(self):
        return self.best_sol

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def choose_starting_solution(self):
    # {
        print(f"SA[{str(self.idnum)}] - Generating a starting solution")
        attempts = 0
        while attempts < 100:
        # {
            print(f"SA[{str(self.idnum)}]. Attempt={str(attempts)}")
            sol = self.generate_solution()  # Generate a solution
            sol, converged = self.evaluate(sol) # Evaluate the solution

            if converged:
                return sol  # Exit the procedure with a valid solution
            attempts += 1   # Update attempts
            print('attempt number', attempts)
        # }
        return None
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.  Evaluate how good a solution is                 '''
    ''' ---------------------------------------------------------- '''
    def evaluate(self, sol):
    # {
        sol, converged = self.evaluate_solution(sol)
        return sol, converged
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Prepare to run the algorithm                     '''
    ''' ---------------------------------------------------------- '''
    def copy_solution(self, sol):
    # {
        copy_sol = copy.deepcopy(sol)
        return copy_sol
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Prepare to run the algorithm                     '''
    ''' ---------------------------------------------------------- '''
    def prepare_to_run(self):
    # {
        self.start = time.time()
        self.no_impr, self.step = 0, 0
        self.t = self.tI  # Set current temperature

        if self.current_sol == None:
            self.current_sol = self.choose_starting_solution()

        # ----------------------------------------------------------
        if self.current_sol is None:
        # {
            print("A feasible starting solution was not generated")
            quit()
        # }
        # _______________________________________________________

        # Add starting solution to the archive
        self.archive.append(self.current_sol)

        # Define best_sol = current_sol
        self.best_sol = self.copy_solution(self.current_sol)

        # Log initial solution and report progress
        print(f"SA[{self.idnum}]. Starting solution: ", self.current_sol.get_obj())
        self.report_progress(self.progress_file)
        self.log_solution("Initial Solution", self.current_sol, file=self.results_file)
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Finish up                                        '''
    ''' ---------------------------------------------------------- '''
    def finalise(self):
    # {
        print(f"Solver[{str(self.idnum)}]. Finalising")
        if self.nb_crit == 1:
            self.log_solution("Final Solution", self.best_sol, file=self.results_file)
            self.log_decision(self.best_sol, file=self.best_file)
        else:
            self.log_archive("Non Dominated Solutions", file=self.archive_file)

        print(f"#Converged={self.converged}; #Not Converged={self.not_converged}", file=self.results_file)
        print(f"#Accepted={self.accepted}; #Not Accepted={self.not_accepted}", file=self.results_file)
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.  Acceptance functions                            '''
    ''' ---------------------------------------------------------- '''
    def accept_change_metrop(self, before, after):
    # {
        """ Evaluate Metropolis function for each objective """
        crits = self.param.criterions

        # Note: crit[1] is the sign and equivalent to crits[i][1]
        rn = np.random.rand()
        accept_i = [np.log(rn) < (crit[1] * (after[i] - before[i]) / self.t)
                    for i, crit in enumerate(crits)]
        return all(accept_i)
    # }

    def accept_change_relative(self, before, after):
    # {
        """ delta > 0 => improvement, delta < 0 => non improvement """

        crits = self.param.criterions
        delta_i = [(crits[i][1] * (after[i] - before[i])) / before[i] for i in range(len(crits))]
        if all(delta > 0 for delta in delta_i): # If all_positive
            return True
        else:
        # {
            if all(delta < 0 for delta in delta_i): # If all negative
                return False
            else:
                ratio = abs(delta_i[0] / delta_i[1])
                return 0.8 <= ratio <= 1.2
        # }
    # }

    # Note: This acceptance strategy does not use the temperature!
    def accept_change_dom(self, before, after):
    # {
        """ Use dominance conditions to accept/reject """
        crits = self.param.criterions

        if dominates(after, before, crits):
            return True # New solution is strictly better so accept it
        if not dominates(before, after, crits):
            return True # Accept as no dominance relationship exists, i.e. new solution is not worse

        # Solution is dominated - Accept 10% of the time
        return (np.random.rand() < 0.10)
    # }

    def accept_change_single(self, before, after):
    # {
        return self.accept_change_metrop(before, after)
    # }

    def accept_change_multi(self, before, after):
    # {
        #return self.accept_change_metrop(before, after)
        # return self.accept_change_relative(before, after)
        return self.accept_change_dom(before, after)
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Restore the best solution                        '''
    ''' ---------------------------------------------------------- '''
    def restore_best(self):
    # {
        self.current_sol = self.copy_solution(self.best_sol)
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Solution improvement process                     '''
    ''' ---------------------------------------------------------- '''
    def improve(self, sol):
    # {
        if np.random.rand() < 0.25:
            sol = self.local_search_distribution(sol, 0)
        else:
        # {
            choices = []
            choices.append(self.local_search_asfeature)
            choices.append(self.local_search_isfeature)
            choices.append(self.local_search_randfeature)
            choice = np.random.choice(choices) # Make a choice
            add = np.random.randint(2)  # Choose to add or remove feature
            sol = choice(sol, 0, add)
        # }
        return sol
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Perturb the solution                             '''
    ''' ---------------------------------------------------------- '''
    def perturb_single(self):
    # {
        self.perturb_solution(self.current_sol) # Perturb current solution
    # }

    def perturb_multi(self):
    #{
        num_par = len(self.archive)         # Number of Pareto-optimal solutions
        chosen = np.random.choice(num_par)  # Choose one solution
        chosen_sol = self.archive[chosen]   # Reference to chosen solution
        self.perturb_solution(chosen_sol)   # Perturb chosen solution
    # }

    def perturb_solution(self, sol):
    # {
        curr_score = [sol.obj(i) for i in range(self.nb_crit)]
        new_sol = self.copy_solution(sol)

        # ~~~~~~~~~~~~
        
        choices = []


        #The idea is we only want to play with the membership feautures first
        if self.optimise_membership is not True:

            choices.append(self.perturb_asfeature)

            if self.param.isvarnames is not None:
                choices.append(self.perturb_isfeature)

            if self.param.asvarnames is not None:
                #Not latent so can add
                #if sol['member_params_spec'] is None:
                choices.append(self.perturb_randfeature)


            if sol['randvars'] is not None:
                choices.append(self.perturb_distribution)

            if self.param.ps_bctrans is None or self.param.ps_bctrans:
                choices.append(self.perturb_bcfeature)

            if self.param.ps_cor is None or self.param.ps_cor:
                choices.append(self.perturb_corfeature)

        if self.optimise_membership:
        # {
            if sol['member_params_spec'] is not None:
                choices.append(self.perturb_member_paramfeature)
        # }
        
        if self.optimise_class:
        # {
            if sol['class_params_spec'] is not None:
                choices.append(self.perturb_class_paramfeature)
        # }
        
       
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Call the chosen perturbation strategy
        choice = np.random.choice(choices)
        new_sol = choice(new_sol)
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Solution evaluation and acceptance handling
        new_sol, converged = self.evaluate(new_sol)
        if converged:
        # {
            new_score = [new_sol.obj(i) for i in range(self.nb_crit)]
            args = (curr_score, new_score)
            if self.accept_change(*args):
            # {
                accd = True
                self.accepted += 1  # Tracking acceptances - Increment counter
                self.current_sol = new_sol  # Set new current solution
                self.update_best(new_sol)  # Update the best solution if necessary
                # Optional: self.log_solution("Perturbation", self.current_sol, file=self.results_file)
            # }
            else:
                accd = False
                self.not_accepted += 1  # Tracking non-acceptances - Increment counter
            # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            self.log_kpi(new_sol, self.debug_file, accd)  # Report to file
        # }
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def update_best(self, sol):
    # {
        if self.nb_crit == 1:
        # {
            if is_better(sol.obj(0), self.best_sol.obj(0), self.param.sign_crit(0)):
                self.best_sol.copy_solution(sol)
                self.no_impr = 0
        # }
        else:
            self.archive = self.update_archive(sol)
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Store non-dominated solutions only               '''
    ''' ---------------------------------------------------------- '''
    def update_archive(self, add_sol):
    # {
        _dominated = []
        for sol in self.archive:
        # {
            if dominates(sol.get_obj(), add_sol.get_obj(), self.param.criterions):
            # {
                # The new solution is dominated by an archive solution
                return self.archive # Return the archive - no need to continue
            # }
            elif dominates(add_sol.get_obj(), sol.get_obj(), self.param.criterions):
            # {
                if sol not in _dominated: _dominated.append(sol)
            # }
        # }

        # Remove all solutions 'add_sol' dominates and add 'add_sol'
        self.archive = [sol for sol in self.archive if sol not in _dominated] + [add_sol]
        #DEBUG: print("The new solution is non-dominated. It dominates ",len(_dominated)," existing solution(s)")

        return self.archive
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' -----------------------------------------------------------'''
    def reset_current_solution(self, size=None):
    # {
        if self.nb_crit == 1:
            self.handle_non_improvement()
        else:
            self.handle_static_archive(size)
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' -----------------------------------------------------------'''
    def handle_non_improvement(self):
    # {
        if is_worse(self.current_sol.obj(0), self.best_sol.obj(0), self.param.sign_crit(0)):
            self.no_impr += 1  # Increment non improvement counter

        if self.no_impr > self.max_no_impr:  # Key step enabling performance
        # {
            print("NO IMPROVEMENT FOR A WHILE. RESTORE BEST SOLUTION.")
            self.restore_best()  # Reinstate the best solution
            self.no_impr = 0  # Reset non improvement counter
        # }
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' -----------------------------------------------------------'''
    def handle_static_archive(self, size):
    # {
        if len(self.archive) == size:
            self.no_impr += 1 # Increment non improvement counter

        if self.no_impr > self.max_no_impr:  # Key step enabling performance
        # {
            print("ARCHIVE STATIC. RESTORE A NON DOMINATED SOLUTION.")
            choice = np.random.randint(len(self.archive))
            self.current_sol = self.archive[choice]
            self.no_impr = 0
        # }
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def synchronize(self):
    # {
        global overall_best_solution
        with lock:
        # {
            if overall_best_solution is None or \
                is_better(self.best_sol.obj(0), overall_best_solution.obj(0), self.param.sign_crit(0)):
                overall_best_solution = self.best_sol  # Update overall best solution
            elif overall_best_solution is not None and \
                is_worse(self.best_sol.obj(0), overall_best_solution.obj(0), self.param.sign_crit(0)):
                self.update_best(overall_best_solution)  # Revise best solution of current SA solver
        # }
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Inner loop of Simulated Annealing algorithm      '''
    ''' ---------------------------------------------------------- '''
    def evaluate_state_changes(self):
    # {
        self.step += 1  # Increment the step variable
        count, size = 0, len(self.archive)
        while (True):
        # {
            self.perturb_function()
            count = count + 1
            if count % 1000 == 0:
                print(f'Iteration at {count}')
            if count >= self.max_iter:
                break
        # }
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # TURNED OFF. UNSATISFACTORY PERFORMANCE SO FAR!
        print(f'step number {self.step}')
        #if (self.step) % 2:
        #    self.best_sol = self.improve(self.best_sol)  # Apply local improvement
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        self.report_progress(self.progress_file)  # Report current state
        self.report_progress()  # Report to the screen
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        self.t = self.rate * self.t  # Reduce the temperature accordingly

        self.reset_current_solution()  # Reset current solution conditionally
    # }

    def frozen(self):
        #if any true return true,
        #else return false
        return (self.get_run_time() > self.max_time or
        self.step > self.max_total_iter or
        self.t < self.tF)
        #return (self.t < self.tF)

    def iterate(self, synch=False):
    # {
        if not self.frozen(): # i.e. t > tF
            self.evaluate_state_changes()
            # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            # Optional - Synchronize with other parallel SA
            #if synch and (self.step % self.comm_int == 0):
            #    self.synchronize()
            # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            self.terminate = False
        else:
            if self.nb_crit == 1:
                self.restore_best() # Reinstate the best solution
            print(f"Solver[{str(self.idnum)}]. Search complete")
            self.terminate = True
        # }
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def report_progress(self, file=None):
    # {
        text = f"SA[{self.idnum}]. Progress @ Step={self.step}; Curr=[{self.current_sol.concatenate_obj()}]; "

        if self.nb_crit == 1:
            text += f"Best=[{self.best_sol.concatenate_obj()}]; "
        else:
            text += f"Archive Size={len(self.archive)}; "

        now = time.time()
        elapsed = now - self.start
        text += f"Elapsed Time={round(elapsed,2)}"

        print(text, file=file)
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. obj[1], obj[2], t/f                                 '''
    ''' ---------------------------------------------------------- '''
    def log_kpi(self, sol, file=None, accept=True):
    # {
        text = sol.concatenate_obj()
        text += ",true," if accept else ",false,"
        text += str(self.step)
        print(text, file=file)
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def log_solution(self, descr, sol, file=None):
    # {
        header = f"Solution: {descr}"
        nchar = len(header)
        line = "_" * nchar
        print(line, file=file)
        print(header, file=file)
        print(line, file=file)
        print("Objectives:", file=file)
        for i in range(self.nb_crit):
        # {
            opt = "Maximise" if self.param.sign_crit(i) == 1 else "Minimise"
            text = f"[{i}]. ({opt}) {self.param.crit(i)} = {round(sol.obj(i), 4)}"
            print(text, file=file)
        # }
        report_model_statistics(sol['model'], file)
        print("", file=file)  # Empty line
        self.log_decision(sol, file=file)
        print("", file=file)  # Empty line
    # }

    def log_decision(self, sol, file=None):
    # {
        print("asvars = ", sol['asvars'], file=file)
        print("isvars = ", sol['isvars'], file=file)
        print("randvars = ", sol['randvars'], file=file)
        print("bcvars = ", sol['bcvars'], file=file)
        print("corvars = ", sol['corvars'], file=file)
        print("bctrans = ", sol['bctrans'], file=file)
        print("asc_ind = ", sol['asc_ind'], file=file)
        print("class_param_spec = ", sol['class_params_spec'], file=file)
        print("member_params_spec = ", sol['member_params_spec'], file=file)
        print("model=", sol['model'].descr, file=file)                                         
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def log_archive(self, descr, file=None):
    # {
        for i, sol in enumerate(self.archive):
        # {
            descr = "Non-Dominated #" + str(i)
            self.log_solution(descr, sol, file=file)
        # }
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function. Outer loop of Simulated Annealing Algorithm      '''
    ''' ---------------------------------------------------------- '''
    def run_search(self):
    # {
        self.prepare_to_run()
        while (True):
        #
            self.iterate()
            if self.terminate: break
        # }
        self.finalise()
    # }


    ''' ---------------------------------------------------------- '''
    ''' Function. Sequential latent class approach                 '''
    ''' ---------------------------------------------------------- '''
    def search_latent_update_single(self, overall_best):
    # {
        sign = self.param.sign_crit(0)     # Shortcut to optimisation sign

        if overall_best is None or is_better(self.best_sol.obj(0), overall_best.obj(0), sign):
            overall_best = self.copy_solution(self.best_sol)

        # Current best solution is worse, so terminate:
        terminate = is_worse(self.best_sol.obj(0), overall_best.obj(0), sign)
        return overall_best, terminate
    # }

    # Terminate if no solution in the archive has a 'class_num' = q
    def search_latent_update_multi(self, q):
    # {
        pareto_class_nums = [sol['class_num'] for sol in self.archive]
        terminate = (max(pareto_class_nums) != q)
        return terminate
    # }

    """def run_search_latent(self, max_classes=5):
    # {
        overall_best_sol = None
        for q in range(1, self.max_classes):
        # {
            print('RunSearchLatent. #classes=', q)
            self.param.latent_class = False if q==1 else True
            self.param.num_classes = q
            # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            # FORCE GENERATION OF A NEW STARTING SOLUTION
            # WITH LATENT CLASS COMPONENTS
            if q == 2:
                del self.current_sol # Delete current solution
                self.current_sol = None
            # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            self.run_search()
            # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            self.current_sol['class_num'] = q
            self.best_sol['class_num'] = q
            if self.nb_crit == 1:
                overall_best_sol, terminate = self.search_latent_update_single(overall_best_sol)
            else:
                terminate = self.search_latent_update_multi(q)
            if terminate: break

            print(f"q={q},terminate={terminate}")
        # }
        print(f"Best solution has {overall_best_sol['class_num']} latent classes")
        self.finalise()
    # }"""

    ####################################################################################
    # THE SEQUENTIAL APPROACH DESCRIBED ABOVE IS NOT BEST PRACTICE.
    # IT MAKES MORE SENSE TO RUN EACH NUMBER OF LATENT CLASSES SEPARATELY, IN PARALLEL
    # AND TO IDENTIFY THE BEST APPROACH AT THE END
    ####################################################################################
    # New & Untested!!!!
    def run_search_latent(self, max_classes=5):
    # {
        init_sol = None
        ctrl = (self.tI, self.tF, self.max_temp_steps, self.max_iter)

        # Define and setup independent solvers
        self.solvers = []
        for q in range(max_classes):
        # {
            self.param.latent_class = False if q == 1 else True
            self.param.num_classes = q
            solver_q = SA(self.param, ctrl, init_sol, idnum=q)
            self.solvers.append(solver_q)

        # }

        with ThreadPoolExecutor() as executor:
            futures = [executor.submit(self.solvers[q].run) for q in range(self.max_classes)]

        for future in as_completed(futures):
            result = future.result()  # This will wait until each task completes

        for q in range(max_classes):
        # {
            self.solvers[q].current_sol['class_num'] = q
            self.solvers[q].best_sol['class_num'] = q
            self.solvers[q].finalise()
        # }

    # }

    ####################################################################################


    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def revise_tI(self, new_tI):
    # {
        self.tI = new_tI
        self.rate = np.exp((1.0 / self.max_temp_steps) * np.log(self.tF/self.tI))
    # }
 # }


''' ----------------------------------------------------------- '''
''' PARALLEL SIMULATED ANNEALING                                '''
''' ----------------------------------------------------------- '''
class PARSA():
# {
    """ Docstring """

    ''' ---------------------------------------------------------- '''
    ''' Function. Constructor                                      '''
    ''' ---------------------------------------------------------- '''
    def __init__(self, param: Parameters, init_sol, ctrl, nthrds=1):
    # {
        self.nthrds = nthrds

        # Define and setup independant solvers
        self.solvers = [SA(param, init_sol, ctrl, idnum=i) for i in range(nthrds)]

        self.choose_custom_tI()  # Optional

        self.comm_int = 1
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def wait(self, futures):
    # {
        for future in as_completed(futures):
            result = future.result()  # This will wait until each task completes
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def run(self):
    # {
        for i in range(self.nthrds):
            self.solvers[i].comm_int = self.comm_int

        with ThreadPoolExecutor() as executor:
            futures = [executor.submit(self.solvers[i].run) for i in range(self.nthrds)]

        self.wait(futures)
        print("PARSA FINISHED!")
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def choose_custom_tI(self, options=None):
    # {
        if options == None or len(options) == 0:
            for _, solver in enumerate(self.solvers):
                solver.revise_tI(np.random.randint(1, 10000))
        else:
            for _, solver in enumerate(self.solvers):
                solver.revise_tI(np.random.choice(options))
    # }
# }


''' ----------------------------------------------------------- '''
''' PARALLEL COOPERATIVE SIMULATED ANNEALING                    '''
''' ----------------------------------------------------------- '''
class PARCOPSA(PARSA):
# {
    """ Docstring """

    ''' ---------------------------------------------------------- '''
    ''' Function. Constructor                                      '''
    ''' ---------------------------------------------------------- '''
    def __init__(self, param: Parameters, init_sol, ctrl, nthrds=1):
    # {
        super().__init__(param, init_sol, ctrl, nthrds)  # Call base class constructor
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def get_best(self):
    # {
        best_at, sign = 0, self.solvers[0].param.sign_crit(0)
        for i in range(1, self.nthrds):
        # {
            obj_i = self.solvers[i].best_sol.get_obj()
            obj_best = self.solvers[best_at].best_sol.get_obj()
            best_at = i if is_better(obj_i, obj_best, sign) else best_at
        # }
        return best_at
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def communicate(self, best_at):
    # {
        best_sol = self.solvers[best_at].best_sol
        for idx, solver in enumerate(self.solvers):
            if idx != best_at:
                solver.update_best(best_sol)
    # }

    ''' ---------------------------------------------------------- '''
    ''' Function.                                                  '''
    ''' ---------------------------------------------------------- '''
    def run(self):
    # {
        with ThreadPoolExecutor(max_workers=self.nthrds) as executor:
        # {
            futures = [executor.submit(self.solvers[i].prepare_to_run) for i in range(self.nthrds)]
            self.wait(futures)
            # ~~~~~~~~~~~~~~~~~~~~~~
            cont, step = True, 0
            while cont:
            # {
                step += 1
                print(f"PARCOPSA. Step {step}")
                # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                futures = [executor.submit(self.solvers[i].iterate) for i in range(self.nthrds)]
                self.wait(futures)
                best_sol = self.get_best()
                self.communicate(best_sol)
                # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                cont = all(not self.solvers[i].terminate for i in range(self.nthrds))
            # }
        # }
        for i in range(self.nthrds):
            self.solvers[i].finalise()
    # }
# }
